"""Module for dealing with student groups."""

from urllib.parse import urlencode
from xml.etree import ElementTree

import requests

import ocflib.account.search as search


_API = {
    'BASE': 'https://studentservices.berkeley.edu/WebServices/StudentGroupServiceV2/Service.asmx',  # noqa
    'SERVICE': {
        'ORGS': 'CalLinkOrganizations',
        'SIGNATORIES_BY_OID': 'CalLinkGroupSignatories',
        'SIGNAT_ACTIVE': 'SignatoriesActiveStudentGroups',
        'SIGNAT_ALL': 'SignatoriesStudentGroups'
    }
}


# TODO: add method(s) for using CalLinkOrganizations service


# TODO: add option to not resolve accounts for speed
def signatories_for_group(oid):
    """Return list of signatories for a group, including name and OCF account.

    >>> signatories_for_group(46187)
    {646431: {'accounts': ['sanjayk'], 'name': 'Sanjay Krishnan'},
     872544: {'accounts': ['daradib'], 'name': 'Mr. Dara Adib'},
     1029873: {'accounts': ['kpengboy'], 'name': 'KEVIN YANG PENG'},
     1031366: {'accounts': ['mattmcal'], 'name': 'Matthew James McAllister'},
     1031553: {'accounts': ['willh'], 'name': 'WILLIAM HO'},
     1032668: {'accounts': ['nickimp'], 'name': 'NICHOLAS DANIEL IMPICCICHE'},
     1034192: {'accounts': ['ckuehl'], 'name': 'CHRISTOPHER B KUEHL'}}
    """
    def parser(root):
        def parse(student):
            uid = int(student.findtext('Username'))

            attrs = search.user_attrs_ucb(uid)
            name = None

            if attrs:
                name = attrs.get('displayName', [None])[0]

            users = search.users_by_calnet_uid(uid)
            return uid, {'name': name, 'accounts': users}

        xml_members = root.findall('Items/Membership')
        return {uid: details for uid, details in map(parse, xml_members)}

    return _get_osl({'organizationId': oid},
                    _API['SERVICE']['SIGNATORIES_BY_OID'], parser)


# TODO: add option to not resolve accounts for speed
def groups_by_student_signat(uid, service=_API['SERVICE']['SIGNAT_ACTIVE']):
    """Return active groups a student is a signatory for.

    >>> groups_by_student_signat(1034192)
    {46187: {'name': 'Open Computing Facility', accounts: ['decal', 'linux']}}
    """
    def parser(root):
        def parse(group):
            oid = int(group.findtext('groupId'))
            return oid, {
                'name': group.findtext('groupName'),
                'accounts':
                    [] if oid == 0 else search.users_by_callink_oid(oid)}

        xml_groups = root.findall('StudentGroupData/StudentGroupDatum')
        return {oid: name for oid, name in map(parse, xml_groups)}

    return _get_osl({'UID': uid}, service, parser)


def groups_by_student_signat_all(uid):
    """Return all (active and inactive) groups a student is a signatory for."""
    return groups_by_student_signat(uid, service=_API['SERVICE']['SIGNAT_ALL'])


def _get_osl(query, service, parser):
    """Query web service for student group information in XML format.

    You should probably use one of the nicer methods instead."""

    url = "{}/{}?{}".format(_API['BASE'], service, urlencode(query))

    r = requests.get(url)
    return _parse_osl(ElementTree.fromstring(r.text), parser)


def _parse_osl(root, parser):
    """Assemble Python dictionaries of groups from XML document"""
    if root.findtext('Succeeded') == 'false':
        try:
            error_reason = root.findtext('Reason')
        except:
            error_reason = 'unknown reason'
        raise Exception("Lookup failed: " + error_reason)

    return parser(root)



import ldap3
from contextlib import contextmanager

import ocflib.constants as constants


@contextmanager
def ldap_connection(host):
    server = ldap3.Server(host, use_ssl=True)
    with ldap3.Connection(server) as connection:
        yield connection


def ldap_ocf():
    return ldap_connection(constants.OCF_LDAP)


def ldap_ucb():
    return ldap_connection(constants.UCB_LDAP)



import ocflib.account.search as search


def get_calnet_names(uid):
    """Returns CalNet LDAP entries relating to names"""

    attrs = search.user_attrs_ucb(uid)
    if attrs:
        return {key: attrs[key]
                for key in ('givenName', 'sn', 'displayname') if attrs[key]}


def name_by_calnet_uid(uid):
    """Returns the name of CalNet person, searched by CalNet UID.

    Returns None on faliure.
    """
    names = get_calnet_names(uid)

    # the name we want to input into our system is "givenName sn"
    # displayName is not necessarily equal to what's printed on Cal 1 Cards
    def get_longest_string(strs):
        return max(strs, key=len)

    if 'givenName' in names and 'sn' in names:
        given_name = get_longest_string(names['givenName'])
        sn = get_longest_string(names['sn'])

        if given_name and sn:
            return "{} {}".format(given_name, sn)
    elif 'displayName' in names:
        display_name = get_longest_string(names["displayName"])

        if display_name:
            return display_name



"""Module containing account management methods, such as password changing,
account creation, etc."""

import base64
import fcntl
import getpass
import paramiko
import pexpect
import socket
from datetime import date

from Crypto.Cipher import PKCS1_OAEP
from Crypto.PublicKey import RSA

import ocflib.account.search as search
import ocflib.account.validators as validators
import ocflib.constants as constants
import ocflib.misc.mail as mail
import ocflib.misc.shell as shell
import ocflib.misc.validators


def change_password_with_staffer(username, password, principal,
                                 admin_password):
    """Change a user's Kerberos password using kadmin and a password, subject
    to username and password validation."""
    validators.validate_username(username)
    validators.validate_password(username, password)

    # try changing using kadmin pexpect
    cmd = "{kadmin_path} -p {principal} cpw {username}".format(
        kadmin_path=shell.escape_arg(constants.KADMIN_PATH),
        principal=shell.escape_arg(principal),
        username=shell.escape_arg(username))

    child = pexpect.spawn(cmd, timeout=10)

    child.expect("{}@OCF.BERKELEY.EDU's Password:".format(username))
    child.sendline(password)
    child.expect("Verify password - {}@OCF.BERKELEY.EDU's Password:"
                 .format(username))
    child.sendline(password)

    # now give admin principal password
    child.expect("{}@OCF.BERKELEY.EDU's Password:".format(principal))
    child.sendline(admin_password)

    child.expect(pexpect.EOF)

    output = child.before.decode('utf8')
    if "Looping detected" in output:
        raise ValueError("Invalid admin password given.")
    elif "kadmin" in output:
        raise ValueError("kadmin Error: {}".format(output))

    _notify_password_change(username)


def change_password_with_keytab(username, password, keytab, principal):
    """Change a user's Kerberos password using a keytab, subject to username
    and password validation."""
    validators.validate_username(username, check_exists=True)
    validators.validate_password(username, password)

    # try changing using kadmin pexpect
    cmd = "{kadmin_path} -K {keytab} -p {principal} cpw {username}".format(
        kadmin_path=shell.escape_arg(constants.KADMIN_PATH),
        keytab=shell.escape_arg(keytab),
        principal=shell.escape_arg(principal),
        username=shell.escape_arg(username))

    child = pexpect.spawn(cmd, timeout=10)

    child.expect("{}@OCF.BERKELEY.EDU's Password:".format(username))
    child.sendline(password)
    child.expect("Verify password - {}@OCF.BERKELEY.EDU's Password:"
                 .format(username))
    child.sendline(password)

    child.expect(pexpect.EOF)

    output = child.before.decode('utf8')
    if "kadmin" in output:
        raise ValueError("kadmin Error: {}".format(output))

    _notify_password_change(username)


def _notify_password_change(username):
    """Send email about a password change."""

    name = search.user_attrs(username)['cn'][0]
    body = """Howdy there {name},

Just a quick heads up that your Open Computing Facility account password was
just reset, hopefully by you.

As a a reminder, your OCF username is: {username}

If you're not sure why this happened, please reply to this email ASAP.

{signature}""".format(name=name,
                      username=username,
                      signature=constants.MAIL_SIGNATURE)

    mail.send_mail_user(username, "[OCF] Account password changed", body)


def trigger_create(ssh_key_path, host_keys_path):
    """Attempt to trigger a create run on the admin server."""

    key = paramiko.RSAKey.from_private_key_file(ssh_key_path)
    ssh = paramiko.SSHClient()
    ssh.load_host_keys(host_keys_path)
    ssh.connect(hostname='admin.ocf.berkeley.edu', username='atool', pkey=key)
    ssh.exec_command('/srv/atool/bin/create')


def encrypt_password(password):
    """Encrypts (not hashes) a user password to be stored on disk while it
    awaits approval.

    Generate the public / private keys with the following code:
    >>> from Crypto.PublicKey import RSA
    >>> key = RSA.generate(2048)
    >>> open("private.pem", "w").write(key.exportKey())
    >>> open("public.pem", "w").write(key.publickey().exportKey())
    """
    # TODO: is there any way we can save the hash instead? this is tricky
    # because we need to stick it in kerberos, but this is bad as-is...
    key = RSA.importKey(open(constants.CREATE_PUBKEY_PATH).read())
    RSA_CIPHER = PKCS1_OAEP.new(key)
    return RSA_CIPHER.encrypt(password)


def queue_creation(full_name, calnet_uid, callink_oid, username, email,
                   password, responsible=None):
    """Queues a user account for creation."""

    # individuals should have calnet_uid, groups should have callink_oid
    if calnet_uid and callink_oid:
        raise ValueError("Only one of calnet_uid or callink_oid may be set.")

    # callink_oid might be 0, which is OK (indicates non-RSO group)
    if not calnet_uid and callink_oid is None:
        raise ValueError("One of calnet_uid or callink_oid must be set.")

    validators.validate_username(username)
    validators.validate_password(username, password)

    if validators.user_exists(username):
        raise ValueError("Username {} is already taken.".format(username))

    if validators.username_queued(username):
        raise ValueError(
            "Username {} is queued for creation.".format(username))

    if validators.username_reserved(username):
        raise ValueError("Username {} is reserved.".format(username))

    full_name = ''.join(c for c in full_name if c.isalpha() or c == ' ')

    if len(full_name) < 3:
        raise ValueError("Full name should be >= 3 characters.")

    if not ocflib.misc.validators.valid_email(email):
        raise ValueError("Email is invalid.")

    # actually create the account
    password = base64.b64encode(encrypt_password(
        password.encode("utf8"))).decode('ascii')

    # TODO: replace this with a better format
    entry_record = [
        username,
        full_name if calnet_uid else '(null)',  # name IF not group
        full_name if not calnet_uid else '(null)',  # name IF group
        email,
        0,
        0 if calnet_uid else 1,
        password,
        calnet_uid or callink_oid,  # university ID
        responsible or '(null)'
    ]

    # same as entry_record but without password
    entry_log = entry_record[:6] + entry_record[7:] + \
        [date.today().isoformat(), getpass.getuser(), socket.getfqdn()]

    # write record to queue and log
    save = (
        (constants.QUEUED_ACCOUNTS_PATH, entry_record),
        (constants.CREATE_LOG_PATH, entry_log)
    )

    for path, entry in save:
        with open(path, 'a') as f:
            fcntl.flock(f, fcntl.LOCK_EX)
            print(':'.join(map(str, entry)), file=f)
            fcntl.flock(f, fcntl.LOCK_UN)



"""Misc utilities for helping with shell commands."""
# shlex.quote is new in Python 3.3
try:  # pragma: no cover
    from shlex import quote as escape_arg  # noqa
except ImportError:  # pragma: no cover
    from pipes import quote as escape_arg  # noqa



from setuptools import setup, find_packages

with open('.version') as f:
    VERSION = f.readline().strip()

setup(
    name='ocflib',
    version=VERSION,
    author="Open Computing Facility",
    author_email='help@ocf.berkeley.edu',
    description="libraries for account and server management",
    url='https://www.ocf.berkeley.edu',
    packages=find_packages(),
    install_requires=(
        'cracklib',
        'dnspython3',
        'ldap3',
        'paramiko',
        'pexpect',
        'requests',
        'pysnmp'
    )
)



import difflib
import pwd
import string
import sys

import cracklib

import ocflib.constants as constants
import ocflib.misc.mail


def validate_username(username, check_exists=False):
    """Validate a username, raising a descriptive exception if problems are
    encountered."""

    if username_reserved(username):
        raise ValueError("Username is reserved")

    if not 3 <= len(username) <= 8:
        raise ValueError("Username must be between 3 and 8 characters")

    if not all(c.islower() for c in username):
        raise ValueError("Username must be all lowercase letters")

    if check_exists and not user_exists(username):
        raise ValueError("Username does not exist")


def validate_password(username, password, strength_check=True):
    """Validate a password, raising a descriptive exception if problems are
    encountered. Optionally checks password strength."""

    if strength_check:
        if len(password) < 8:
            raise ValueError("Password must be at least 8 characters")

        s = difflib.SequenceMatcher()
        s.set_seqs(password, username)

        if s.ratio() > 0.6:
            raise ValueError("Password is too similar to username")

        try:
            cracklib.FascistCheck(password)
        except ValueError as e:
            raise ValueError("Password problem: {}".format(e))

    # sanity check; note we don't use string.whitespace since we don't want
    # tabs or newlines
    allowed = string.digits + string.ascii_letters + string.punctuation + ' '

    if not all(c in allowed for c in password):
        raise ValueError("Password contains forbidden characters")


def user_exists(username):
    try:
        pwd.getpwnam(username)
    except KeyError:
        return False
    else:
        return True


def username_reserved(username):
    if username.startswith('ocf'):
        return True

    # sanity check: make sure no local users share the username
    with open('/etc/passwd') as f:
        if any(line.startswith(username + ':') for line in f):
            print(
                "WARNING: Username {} rejected based on /etc/passwd!"
                .format(username),
                file=sys.stderr)
            ocflib.misc.mail.send_problem_report(
                """Username {} rejected based on /etc/passwd. It should be \
added to RESERVED_USERNAMES for consistency across \
servers!""".format(username))
            return True

    return username in constants.RESERVED_USERNAMES


def username_queued(username):
    """Returns if the username has already been requested and is queued to be
    created."""
    with open(constants.QUEUED_ACCOUNTS_PATH) as f:
        return any(line.startswith(username + ":") for line in f)
